# -*- coding: utf-8 -*-
import re

from pylons import config
import formalchemy
from formalchemy import helpers as h
from sqlalchemy.util import OrderedDict
from pylons.i18n import _, ungettext, N_, gettext

from common import ResourcesField, TagField, ExtrasField, GroupSelectField, package_name_validator
from builder import FormBuilder
import ckan.model as model
import ckan.lib.helpers
from ckan.lib.helpers import literal

__all__ = ['prettify', 'build_package_form', 'get_standard_fieldset']


def prettify(field_name):
    field_name = re.sub('(?<!\w)[Uu]rl(?!\w)', 'URL', field_name.replace('_', ' ').capitalize())
    return _(field_name.replace('_', ' '))

def build_package_form(is_admin=False, user_editable_groups=None, **params):
    builder = FormBuilder(model.Package)

    # Extra fields
    builder.add_field(GroupSelectField('groups', allow_empty=True, user_editable_groups=user_editable_groups))
    builder.add_field(ResourcesField('resources', hidden_label=True))
    builder.add_field(TagField('tags'))
    builder.add_field(ExtrasField('extras', hidden_label=True))

    # Labels and instructions
    builder.set_field_text(
        'title',
        instructions='Le titre du jeu de donnees.',
        further_instructions='Indiquez un titre bref decrivant les donnees. Vous pourrez decrire plus clairement les donnees dans le champ Notes. Veuillez eviter les points de suspension.',
    )
    builder.set_field_text(
        'name', 
        '%s %s' % (
            _('Name'),
            _('(required)')
        ), 
        hints=literal(_("<strong>Unique identifier</strong> for package.<br/>2+ chars, lowercase, using only 'a-z0-9' and '-_'"))
    )
    builder.set_field_text(
        'version',
        instructions='Un nombre indiquant la version du jeu de donnees (si possible). ex : 1.2.0',
    )
    builder.set_field_text(
        'url',
        instructions='Il s\'agit generalement de l\'URL du site hebergeant les donnees, pas celle des donnees elles-memes.',
        hints='ex : http://www.example.com/growth-figures.html',
    )
    builder.set_field_text(
        'author',
        instructions='La personne a contacter pour toute question sur ces donnees en particulier.',
    )
    builder.set_field_text(
        'author_email',
    )
    builder.set_field_text(
        'license_id',
        _('License'),
        instructions='La licence sous laquelle les donnees sont publiees.',
    )
    builder.set_field_text(
        'tags', 
        _('Tags'), 
        _('(space separated list)')
    )
    builder.set_field_text(
        'resources',
        instructions='Les fichiers contenant les donnees ou l\'adresse de l\'API permettant d\'y acceder.',
        further_instructions=literal('<br />&nbsp;Vous pouvez enregistrer autant d\'entrees que necessaire. Par exemplee, si les donnees sont fournies dans differents formats, ou divises en differents lieux ou periodes, chaque fichier est une differente \'resource\' qui doit etre decrite differemment. Elles apparaitront ensuite toutes ensemble sur la page du jeu de donnees.<br /><br /> <b>URL :</b>Il s\'agit du lien direct vers les donnees sur Internet - en cliquant sur ce lien dans un navigateur, l\'utilisateur teelechargera directement le jeu de donnees complet. Notez qu\'aucun jeu de donnees n\est stocke sur ce site, mais par les editeurs des donnees. Alternativemment l\'URL peut diriger vers un serveur d\'API tel qu\'un terminal SPARQL ou un service JSON-P.<br /> <b>Format :</b>Ce champ indique le type de format de fichier dans lequel les donnees sont fournies. <br /><b>Description :</b>Toute information utile que vous souhaitez ajouter a la ressource.<br />'),  
        hints='Exemples de formats : CSV | RDF | XML | XBRL | SDMX | HTML+RDFa | Autre...'
    )
    builder.set_field_text(
        'notes', 
        _('Notes'), 
        instructions='Description generale des donnees',
        further_instructions='Cette description etant tronquee dans certaines vues des jeux de donnees, vous etes invites a commencer par une courte phrase decrivant brievement les donnees avant de fournir des informations plus detaillees.',
        hints=literal(_('You can use <a href="http://daringfireball.net/projects/markdown/syntax">Markdown formatting</a> here.'))
    )

    # Options/settings
    builder.set_field_option('name', 'validate', package_name_validator)
    builder.set_field_option('license_id', 'dropdown', {'options':[('', None)] + model.Package.get_license_options()})
    builder.set_field_option('state', 'dropdown', {'options':model.State.all})
    builder.set_field_option('notes', 'textarea', {'size':'60x15'})

    # Layout
    field_groups = OrderedDict([
        (_('Basic information'), ['title', 'name', 'version', 'url',
                               'notes', 'tags']),
        (_('Resources'), ['resources']),
        (_('Groups'), ['groups']),
        (_('Detail'), ['author', 'author_email',
                    'maintainer', 'maintainer_email',
                    'license_id']),
        (_('Extras'), ['extras']),
        ])
    if is_admin:
        field_groups[_('Detail')].append('state')
    builder.set_displayed_fields(field_groups)
    builder.set_label_prettifier(prettify)
    return builder
    # Strings for i18n:
    [ _('Title'), _('Name'), _('Version'), _('URL'),
     _('Notes'), _('Resources'), _('Author'), _('Author email'), _('Maintainer'),
     _('Maintainer email'), _('License'), _('Tags'), _('Extras'), _('State')]

def get_standard_fieldset(is_admin=False, user_editable_groups=None, **kwargs):
    '''Returns the package fieldset (optionally with admin fields)'''

    return build_package_form(is_admin=is_admin, user_editable_groups=user_editable_groups).get_fieldset()
